from __future__ import annotations  # noqa: F407

import json
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # error in sarus_data_spec.typing

Transform = Tuple[str, Dict[str, Any]]


def from_v2(dataset: Any) -> Any:
    """Compatibility layer for the preprocessing SDK and TF.

    Builder for the class converting a sarus.Dataset into an object
    compatible with the previous SDK API of sarus.Dataset so that TenssorFlow
    calls keep working.
    """
    dataset_cls: Any = type(dataset)

    class CompatibleDataset(dataset_cls):
        """A class keeping the old Dataset API to maintain compatibility."""

        def _add_transform(
            self, transform: Tuple[str, Dict]
        ) -> CompatibleDataset:
            """Create a new dataset with an additional transform."""
            return CompatibleDataset(
                id=self.id,
                client=self.client,
                dataspec=self._dataspec,
                type_metadata=json.dumps(self.type_metadata),
                human_description=self.human_description,
                marginals=json.dumps(self.marginals),
                policy=self.policy,
                synthetic=self._synthetic,
                transforms=self.transforms + [transform],
            )

    return CompatibleDataset(
        id=dataset.id,
        client=dataset.client,
        dataspec=dataset._dataspec,
        type_metadata=json.dumps(dataset.type_metadata),
        human_description=dataset.human_description,
        marginals=json.dumps(dataset.marginals),
        policy=dataset.policy,
        synthetic=dataset._synthetic,
        transforms=dataset.transforms,
    )


class Dataset:
    """A class allowing to manipulate a Sarus dataset as a Tensorflow dataset.

    The `sarus.tensorflow.Dataset` class allows to manipulate a Sarus Dataset as
    if it were a `tensorflow.data.Dataset` instance.

    Similarly to a Sarus Dataset, a `sarus.tensorflow.Dataset` uses synthetic
    data locally, which enables interactions with the data while preserving
    privacy. This allows data practitionners to design their processing
    pipelines and ML models locally as if they were working with a tensorflow
    dataset.

    Under the hood, the class memorizes the transformations that will be applied
    to the private dataset when training it remotely with differential privacy.

    Note: Only pure processing functions should be used for DP reasons.
    """

    def __init__(
        self,
        dataset: Any,
        max_download_size: Optional[int] = None,
        original: bool = False,
    ) -> None:
        self.dataset = from_v2(dataset)
        self.max_download_size = max_download_size
        self.original = original

    def map(
        self,
        map_func: Callable,
        num_parallel_calls: Optional[int] = None,
        deterministic: Optional[bool] = None,
    ) -> Dataset:
        """Map `map_func` across the elements of the dataset.

        This transformation applies `map_func` to each element of this
        dataset, and returns a new dataset containing the transformed
        elements, in the same order as they appeared in the input. `map_func`
        can be used to change both the values and the structure of a dataset's
        elements. Supported structure constructs are documented `here
        <https://www.tensorflow.org/guide/data#dataset_structure>`_.
        For example, `map` can be used for adding 1 to each element, or
        projecting a subset of element components.

        >>> sarus_tf_dataset = sarus_dataset.as_tensorflow()  # ==> [ 1, 2, 3, 4, 5 ]
        >>> sarus_tf_dataset = sarus_tf_dataset.map(lambda x: x + 1)
        >>> list(sarus_tf_dataset.as_numpy_iterator())
        [2, 3, 4, 5, 6]

        The input signature of `map_func` is determined by the structure of each
        element in this dataset. For more information on `map` please refer to the
        `Tensorflow documentation <https://www.tensorflow.org/api_docs/python/tf/data/Dataset#map>`_.

        >>> sarus_tf_dataset = sarus_dataset.as_tensorflow()  # ==> range(5)
        >>> # `map_func` takes a single argument of type `tf.Tensor` with the same
        >>> # shape and dtype.
        >>> result = dataset.map(lambda x: x + 1)

        >>> # Each element is a tuple containing two `tf.Tensor` objects.
        >>> # elements = [(1, "foo"), (2, "bar"), (3, "baz")]
        >>> # `map_func` takes two arguments of type `tf.Tensor`. This function
        >>> # projects out just the first component.
        >>> sarus_tf_dataset = sarus_tf_dataset.map(lambda x_int, y_str: x_int)
        >>> list(sarus_tf_dataset.as_numpy_iterator())
        [1, 2, 3]

        >>> # Each element is a dictionary mapping strings to `tf.Tensor` objects.
        >>> # elements =  ([{"a": 1, "b": "foo"},
        ... #              {"a": 2, "b": "bar"},
        ... #              {"a": 3, "b": "baz"}])
        >>> # `map_func` takes a single argument of type `dict` with the same keys
        >>> # as the elements.
        >>> sarus_tf_dataset = sarus_tf_dataset.map(lambda d: str(d["a"]) + d["b"])

        Args:

        map_func:
            A function mapping a dataset element to another dataset element.

        num_parallel_calls:
            (Optional) A `tf.int64` scalar `tf.Tensor`,
            representing the number elements to process asynchronously in parallel.
            If not specified, elements will be processed sequentially. If the value
            `tf.data.AUTOTUNE` is used, then the number of parallel
            calls is set dynamically based on available CPU.

        deterministic:
            (Optional) When `num_parallel_calls` is specified, if this
            boolean is specified (`True` or `False`), it controls the order in which
            the transformation produces elements. If set to `False`, the
            transformation is allowed to yield elements out of order to trade
            determinism for performance. If not specified, the
            `tf.data.Options.experimental_deterministic` option
            (`True` by default) controls the behavior.

        Returns:
            sarus.tensorflow.Dataset: A `sarus.tensorflow.Dataset`.
        """
        transform = (
            "map",
            {
                "map_func": map_func,
                "num_parallel_calls": num_parallel_calls,
                "deterministic": deterministic,
            },
        )
        new_dataset = self.dataset._add_transform(transform)
        return Dataset(new_dataset, self.max_download_size, self.original)

    def unbatch(self) -> Dataset:
        """Split elements of a dataset into multiple elements.

        For example, if elements of the dataset are shaped `[B, a0, a1, ...]`,
        where `B` may vary for each input element, then for each element in the
        dataset, the unbatched dataset will contain `B` consecutive elements
        of shape `[a0, a1, ...]`.

        >>> sarus_tf_dataset = sarus_dataset.as_tensorflow()
        >>> print(next(iter(sarus_tf_dataset)).take(3))
        [ [1, 2, 3], [1, 2], [1, 2, 3, 4] ]
        >>> sarus_tf_dataset = sarus_tf_dataset.unbatch()
        >>> print(next(iter(sarus_tf_dataset)).take(9))
        [1, 2, 3, 1, 2, 1, 2, 3, 4]

        Note: `unbatch` requires a data copy to slice up the batched tensor into
        smaller unbatched tensors. When optimizing performance, try to avoid
        unnecessary usage of `unbatch`.

        Returns:
            sarus.tensorflow.Dataset: A `sarus.tensorflow.Dataset`.
        """
        transform: Transform = ("unbatch", {})
        new_dataset = self.dataset._add_transform(transform)
        return Dataset(new_dataset, self.max_download_size, self.original)

    def batch(
        self,
        batch_size: int,
        drop_remainder: bool = True,
        num_parallel_calls: Optional[int] = None,
        deterministic: Optional[bool] = None,
    ) -> Dataset:
        """Combine consecutive elements of this dataset into batches.

        >>> sarus_tf_dataset = sarus_dataset.as_tensorflow()
        >>> sarus_tf_dataset = sarus_tf_dataset.batch(3)

        The components of the resulting element will have an additional outer
        dimension, which will be `batch_size`.
        Currently, `drop_remainder`can only take True value for now.
        If you set the value to False, it will be reset to True.

        For remote private training, the dataset can be batched only once.
        Better performance is obtained if the preprocessing is applied after
        batching the dataset once.

        >>> sarus_tf_dataset.map(preprocess).batch(64)  # poor performance
        >>> sarus_tf_dataset.batch(64).map(preprocess)  # optimal performance
        >>> sarus_tf_dataset.batch(32).batch(32)  # error, not allowed
        >>> sarus_tf_dataset.map(preprocess)  # allowed but slow

        Args:

        batch_size:
            A `tf.int64` scalar `tf.Tensor`, representing the number of
            consecutive elements of this dataset to combine in a single batch.

        drop_remainder:
            (Optional) A `tf.bool` scalar `tf.Tensor`, representing whether the
            last batch should be dropped in the case it has fewer than
            `batch_size` elements; Currently, only a value of True is accepted
            for now. If you set the value to False, it will be reset to True.

        num_parallel_calls:
            (Optional) A `tf.int64` scalar `tf.Tensor`, representing the number
            of batches to compute asynchronously in parallel. If not specified,
            batches will be computed sequentially. If the value
            `tf.data.AUTOTUNE` is used, then the number of parallel calls is set
            dynamically based on available resources.

        deterministic:
            (Optional) When `num_parallel_calls` is specified, if this boolean
            is specified (`True` or `False`), it controls the order in which the
            transformation produces elements. If set to `False`, the
            transformation is allowed to yield elements out of order to trade
            determinism for performance. If not specified, the
            `tf.data.Options.experimental_deterministic` option (`True` by
            default) controls the behavior.

        Returns:
            sarus.tensorflow.Dataset: A `sarus.tensorflow.Dataset`.
        """
        # TODO `drop_remainder` = False not supported yet
        if drop_remainder is False:
            logging.info(
                "Changing `drop_remainder` to True as setting it to False "
                "is not supported yet."
            )
            drop_remainder = True

        transform = (
            "batch",
            {
                "batch_size": batch_size,
                "drop_remainder": drop_remainder,
                "num_parallel_calls": num_parallel_calls,
                "deterministic": deterministic,
            },
        )
        new_dataset = self.dataset._add_transform(transform)
        return Dataset(new_dataset, self.max_download_size, self.original)

    def filter(self, predicate: Callable) -> Dataset:
        """Filter the dataset according to `predicate`.

        >>> sarus_tf_dataset = sarus_dataset.as_tensorflow() # ==> [1, 2, 3]
        >>> sarus_tf_dataset = sarus_tf_dataset.filter(lambda x: x < 3)
        >>> list(sarus_tf_dataset.as_numpy_iterator())
        [1, 2]

        >>> # `tf.math.equal(x, y)` is required for equality comparison
        >>> def filter_fn(x):
        ...   return tf.math.equal(x, 1)
        >>> sarus_tf_dataset = sarus_tf_dataset.filter(filter_fn)
        >>> list(sarus_tf_dataset.as_numpy_iterator())
        [1]

        Args:
        predicate:
            A function mapping a dataset element to a boolean.

        Returns:
            sarus.tensorflow.Dataset: The `sarus.tensorflow.Dataset` containing
            the elements of the dataset for which `predicate` is `True`.
        """
        transform = ("filter", {"predicate": predicate})
        new_dataset = self.dataset._add_transform(transform)
        return Dataset(new_dataset, self.max_download_size, self.original)

    def _tensorflow(self) -> tf.data.Dataset:
        ds = self.dataset._synthetic_as_tf_dataset(
            batch_size=1,
            rows_number=self.max_download_size,
            original=self.original,
        )
        return _apply_transforms(ds, self.dataset.transforms)

    def __iter__(self) -> Iterator:
        return self._tensorflow().__iter__()


def _apply_transforms(
    ds: tf.data.Dataset, transforms: List[Transform]
) -> tf.data.Dataset:
    for name, params in transforms:
        if name == "map":
            ds = ds.map(**params)
        elif name == "unbatch":
            ds = ds.unbatch(**params)
        elif name == "batch":
            ds = ds.batch(**params)
        elif name == "filter":
            ds = ds.filter(**params)
        elif name == "split":
            size = params["end"] - params["start"]
            ds = ds.skip(params["start"]).take(size)
    return ds



from __future__ import annotations

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import sklearn.decomposition as decomposition  # noqa: F401
    from sklearn.decomposition import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class PCA(DataSpecWrapper[decomposition.PCA]):
    @sarus_init("sklearn.SK_PCA")
    def __init__(
        self,
        n_components=None,
        *,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()



from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import sklearn.pipeline as sk_pipeline
    from sklearn.pipeline import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class Pipeline(DataSpecWrapper[sk_pipeline.Pipeline]):
    @sarus_init("sklearn.SK_PIPELINE")
    def __init__(self, steps, *, memory=None, verbose=False):
        ...

    @sarus_method("sklearn.SK_PIPELINE_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()



"""Sarus Pandas package documentation."""
# flake8: noqa
from pandas import *

from sarus.utils import register_ops

from .core import *
from .dataframe import DataFrame, Series
from .io import *

register_ops()



from __future__ import annotations

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import sklearn.ensemble as ensemble  # noqa: F401
    from sklearn.ensemble import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class AdaBoostClassifier(DataSpecWrapper[ensemble.AdaBoostClassifier]):
    @sarus_init("sklearn.SK_ADABOOST_CLASSIFIER")
    def __init__(
        self,
        base_estimator=None,
        *,
        n_estimators=50,
        learning_rate=1.0,
        algorithm="SAMME.R",
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class AdaBoostRegressor(DataSpecWrapper[ensemble.AdaBoostRegressor]):
    @sarus_init("sklearn.SK_ADABOOST_REGRESSOR")
    def __init__(
        self,
        base_estimator=None,
        *,
        n_estimators=50,
        learning_rate=1.0,
        loss="linear",
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class BaggingClassifier(DataSpecWrapper[ensemble.BaggingClassifier]):
    @sarus_init("sklearn.SK_BAGGING_CLASSIFIER")
    def __init__(
        self,
        base_estimator=None,
        n_estimators=10,
        *,
        max_samples=1.0,
        max_features=1.0,
        bootstrap=True,
        bootstrap_features=False,
        oob_score=False,
        warm_start=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class BaggingRegressor(DataSpecWrapper[ensemble.BaggingRegressor]):
    @sarus_init("sklearn.SK_BAGGING_REGRESSOR")
    def __init__(
        self,
        base_estimator=None,
        *,
        n_estimators=50,
        learning_rate=1.0,
        loss="linear",
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class ExtraTreesClassifier(DataSpecWrapper[ensemble.ExtraTreesClassifier]):
    @sarus_init("sklearn.SK_EXTRA_TREES_CLASSIFIER")
    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="gini",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=False,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0.0,
        max_samples=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class ExtraTreesRegressor(DataSpecWrapper[ensemble.ExtraTreesRegressor]):
    @sarus_init("sklearn.SK_EXTRA_TREES_REGRESSOR")
    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=False,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class GradientBoostingClassifier(
    DataSpecWrapper[ensemble.GradientBoostingClassifier]
):
    @sarus_init("sklearn.SK_GRADIENT_BOOSTING_CLASSIFIER")
    def __init__(
        self,
        *,
        loss="log_loss",
        learning_rate=0.1,
        n_estimators=100,
        subsample=1.0,
        criterion="friedman_mse",
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_depth=3,
        min_impurity_decrease=0.0,
        init=None,
        random_state=None,
        max_features=None,
        verbose=0,
        max_leaf_nodes=None,
        warm_start=False,
        validation_fraction=0.1,
        n_iter_no_change=None,
        tol=0.0001,
        ccp_alpha=0.0,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class GradientBoostingRegressor(
    DataSpecWrapper[ensemble.GradientBoostingRegressor]
):
    @sarus_init("sklearn.SK_GRADIENT_BOOSTING_REGRESSOR")
    def __init__(
        self,
        *,
        loss="squared_error",
        learning_rate=0.1,
        n_estimators=100,
        subsample=1.0,
        criterion="friedman_mse",
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_depth=3,
        min_impurity_decrease=0.0,
        init=None,
        random_state=None,
        max_features=None,
        alpha=0.9,
        verbose=0,
        max_leaf_nodes=None,
        warm_start=False,
        validation_fraction=0.1,
        n_iter_no_change=None,
        tol=0.0001,
        ccp_alpha=0.0,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class IsolationForest(DataSpecWrapper[ensemble.IsolationForest]):
    @sarus_init("sklearn.SK_ISOLATION_FOREST")
    def __init__(
        self,
        *,
        n_estimators=100,
        max_samples="auto",
        contamination="auto",
        max_features=1.0,
        bootstrap=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class RandomForestClassifier(DataSpecWrapper[ensemble.RandomForestClassifier]):
    @sarus_init("sklearn.SK_RANDOM_FOREST_CLASSIFIER")
    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="gini",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0.0,
        max_samples=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...

    @sarus_method("sklearn.SK_PREDICT")
    def predict(self, X):
        ...


class RandomForestRegressor(DataSpecWrapper[ensemble.RandomForestRegressor]):
    @sarus_init("sklearn.SK_RANDOM_FOREST_REGRESSOR")
    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=False,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class RandomTreesEmbedding(DataSpecWrapper[ensemble.RandomTreesEmbedding]):
    @sarus_init("sklearn.SK_RANDOM_TREES_EMBEDDING")
    def __init__(
        self,
        n_estimators=100,
        *,
        max_depth=5,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        sparse_output=True,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class StackingClassifier(DataSpecWrapper[ensemble.StackingClassifier]):
    @sarus_init("sklearn.SK_STACKING_CLASSIFIER")
    def __init__(
        self,
        estimators,
        final_estimator=None,
        *,
        cv=None,
        stack_method="auto",
        n_jobs=None,
        passthrough=False,
        verbose=0,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class StackingRegressor(DataSpecWrapper[ensemble.StackingRegressor]):
    @sarus_init("sklearn.SK_STACKING_REGRESSOR")
    def __init__(
        self,
        estimators,
        final_estimator=None,
        *,
        cv=None,
        stack_method="auto",
        n_jobs=None,
        passthrough=False,
        verbose=0,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class VotingClassifier(DataSpecWrapper[ensemble.VotingClassifier]):
    @sarus_init("sklearn.SK_VOTING_CLASSIFIER")
    def __init__(
        self,
        estimators,
        *,
        voting="hard",
        weights=None,
        n_jobs=None,
        flatten_transform=True,
        verbose=False,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class VotingRegressor(DataSpecWrapper[ensemble.VotingRegressor]):
    @sarus_init("sklearn.SK_VOTING_REGRESSOR")
    def __init__(
        self,
        estimators,
        *,
        weights=None,
        n_jobs=None,
        verbose=False,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class HistGradientBoostingClassifier(
    DataSpecWrapper[ensemble.HistGradientBoostingClassifier]
):
    @sarus_init("sklearn.SK_HIST_GRADIENT_BOOSTING_CLASSIFIER")
    def __init__(
        self,
        loss="log_loss",
        *,
        learning_rate=0.1,
        max_iter=100,
        max_leaf_nodes=31,
        max_depth=None,
        min_samples_leaf=20,
        l2_regularization=0.0,
        max_bins=255,
        categorical_features=None,
        monotonic_cst=None,
        warm_start=False,
        early_stopping="auto",
        scoring="loss",
        validation_fraction=0.1,
        n_iter_no_change=10,
        tol=1e-07,
        verbose=0,
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method(
        "sklearn.SK_FIT",
        inplace=True,
    )
    def fit(self, X, y=None):
        ...


class HistGradientBoostingRegressor(
    DataSpecWrapper[ensemble.HistGradientBoostingRegressor]
):
    @sarus_init("sklearn.SK_HIST_GRADIENT_BOOSTING_REGRESSOR")
    def __init__(
        self,
        estimators,
        *,
        weights=None,
        n_jobs=None,
        verbose=False,
        _dataspec=None,
    ):
        ...

    @sarus_method(
        "sklearn.SK_FIT",
        inplace=True,
    )
    def fit(self, X, y=None):
        ...


register_ops()



import logging
import os
import time
import traceback
import typing as t

import pyarrow as pa
import pyarrow.parquet as pq
from sarus_data_spec import typing as st
from sarus_data_spec.constants import CACHE_PATH, TO_PARQUET_TASK
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.base import Base
from sarus_data_spec.manager.computations.local.base import LocalComputation
from sarus_data_spec.status import DataSpecErrorStatus, error, ready

from sarus.manager.arrow_remote import ToArrowComputationOnServer

logger = logging.getLogger(__name__)

BATCH_SIZE = 10000


class ToParquetComputation(LocalComputation[str]):
    """Class responsible for handling the caching in parquet of a dataset. It wraps a ToArrowComputation to get the iterator."""

    task_name = TO_PARQUET_TASK

    def __init__(
        self, computing_manager: Base, arrow_remote: ToArrowComputationOnServer
    ) -> None:
        super().__init__(computing_manager)
        self.remote_arrow = arrow_remote

    async def prepare(self, dataspec: st.DataSpec) -> None:
        logger.debug(f"STARTING LOCAL TO_PARQUET {dataspec.uuid()}")
        start = time.perf_counter()
        try:
            if self.computing_manager().is_computation_remote(dataspec):
                iterator = await self.remote_arrow.task_result(
                    dataspec, batch_size=BATCH_SIZE
                )
            else:
                iterator = await self.computing_manager().async_to_arrow_op(
                    dataset=t.cast(Dataset, dataspec), batch_size=BATCH_SIZE
                )

            batches = [batch async for batch in iterator]
            pq.write_table(
                table=pa.Table.from_batches(batches),
                where=self.cache_path(dataspec=dataspec),
                version="2.6",
            )
        except DataSpecErrorStatus as exception:
            error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(exception.relaunch),
                },
            )
            raise
        except Exception:
            error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(False),
                },
            )
            raise DataSpecErrorStatus((False, traceback.format_exc()))
        else:
            end = time.perf_counter()
            logger.debug(
                f"FINISHED LOCAL TO_PARQUET {dataspec.uuid()} ({end-start:.2f}s)"
            )
            ready(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=TO_PARQUET_TASK,
                properties={CACHE_PATH: self.cache_path(dataspec)},
            )

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> str:
        """Returns the cache_path."""
        return properties[CACHE_PATH]

    def cache_path(self, dataspec: st.DataSpec) -> str:
        """Returns the path where to cache the dataset."""
        return os.path.join(
            dataspec.manager().parquet_dir(), f"{dataspec.uuid()}.parquet"
        )



from __future__ import annotations

import json
import logging
import os
import tarfile
import tempfile
import textwrap
from typing import Any, Callable, Container, Dict, List, Optional, Tuple, Union

import cloudpickle
import numpy as np

from sarus.legacy.tensorflow import Dataset

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # error in sarus_data_spec.typing


Transform = Tuple[str, Dict[str, Any]]
Data = Any


class Model(tf.keras.Model):
    """A class similar to a keras Model and allowing private remote training.

    The sarus.keras.Model class is a wrapper around the
    `tensorflow.keras.Model` class. This class differs from its parent only on
    the `fit` method. The `fit` method accepts a `target_epsilon`
    (Differential Privacy parameter).

        - If the specified `target_epsilon` equals to 0, then the `Model` class launches a standard keras training on the synthetic data.

        - If the specified `target_epsilon` is strictly greater than 0, then the `Model` class calls the Sarus API to launch a remote private training with `target_epsilon` the maximum privacy budget to assign to the differentially-private model fitting.

        - If None, a default `target_epsilon` specific to the current user and access rule is used.
            Default target epsilon is 0 if the access is a Differentially-Private access with
            per-user or per-group limit; default value equals to the per-query limit if the access is
            a Differentially-Private access with a per-query limit only. Meaning Sarus maximizes the result
            accuracy in the given privacy constraints. See user documentation to know more.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(Model, self).__init__(self, *args, **kwargs)

    def predict(
        self,
        x: Data,
        batch_size: Optional[int] = None,
        verbose: int = 0,
        steps: Optional[int] = None,
        callbacks: Optional[List[tf.keras.callbacks.Callback]] = None,
        max_queue_size: int = 10,
        workers: int = 1,
        use_multiprocessing: bool = False,
    ) -> np.ndarray:
        """Generate output predictions for the input samples.

        Computation is done in batches. This method is designed for performance
        in large scale inputs. For small amount of inputs that fit in one batch,
        directly using `__call__` is recommended for faster execution, e.g.,
        `model(x)`, or `model(x, training=False)` if you have layers such as
        `tf.keras.layers.BatchNormalization` that behaves differently during
        inference. Also, note the fact that test loss is not affected by
        regularization layers like noise and dropout.

        Args:

        x: Input samples.
            It could be:

            - A Numpy array (or array-like), or a list of arrays
                in case the model has multiple inputs.
            - A TensorFlow tensor, or a list of tensors
                in case the model has multiple inputs.
            - A `tf.data` dataset.
            - A `sarus.tensorflow.Dataset` dataset.
            - A generator or `keras.utils.Sequence` instance.

        batch_size (int, optional):
            Number of samples per batch.
            If unspecified, `batch_size` will default to 32.
            Do not specify the `batch_size` if your data is in the
            form of dataset, `sarus.tensorflow.Dataset`, generators, or
            `keras.utils.Sequence` instances (since they generate batches).

        verbose (int):
            Verbosity mode, 0 or 1.

        steps (int):
            Total number of steps (batches of samples)
            before declaring the prediction round finished.
            Ignored with the default value of `None`. If x is a `tf.data`
            dataset and `steps` is None, `predict` will
            run until the input dataset is exhausted.

        callbacks: List of `keras.callbacks.Callback` instances.
            List of callbacks to apply during prediction. See
            `Tensorflow callbacks documentation
            <https://www.tensorflow.org/api_docs/python/tf/keras/callbacks>`_.

        max_queue_size (int):
            Used for generator or `keras.utils.Sequence`
            input only. Maximum size for the generator queue.
            If unspecified, `max_queue_size` will default to 10.

        workers (int):
            Used for generator or `keras.utils.Sequence` input
            only. Maximum number of processes to spin up when using
            process-based threading. If unspecified, `workers` will default
            to 1.

        use_multiprocessing (bool):
            Used for generator or
            `keras.utils.Sequence` input only. If `True`, use process-based
            threading. If unspecified, `use_multiprocessing` will default to
            `False`. Note that because this implementation relies on
            multiprocessing, you should not pass non-picklable arguments to
            the generator as they can't be passed easily to children
            processes.

        See the discussion of `Unpacking behavior for iterator-like inputs` for
        `tensorflow.keras.Model.fit`.
        Note that `Model.predict` uses the same interpretation rules
        as `Model.fit` and `Model.evaluate`, so inputs must be unambiguous for
        all three methods.
        `Model.predict` is not yet supported with
        `tf.distribute.experimental.ParameterServerStrategy`.

        Returns:
            Numpy array(s) of predictions.

        Raises:
            RuntimeError:
                If `model.predict` is wrapped in `tf.function`.

            ValueError:
                In case of mismatch between the provided
                input data and the model's expectations,
                or in case a stateful model receives a number of samples
                that is not a multiple of the batch size.
        """
        if isinstance(x, Dataset):
            x = x._tensorflow()
        return super().predict(
            x=x,
            batch_size=batch_size,
            verbose=verbose,
            steps=steps,
            callbacks=callbacks,
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
        )

    def evaluate(
        self,
        x: Data = None,
        y: Data = None,
        batch_size: Optional[int] = None,
        verbose: int = 1,
        sample_weight: Optional[np.ndarray] = None,
        steps: Optional[int] = None,
        callbacks: Optional[List[tf.keras.callbacks.Callback]] = None,
        max_queue_size: int = 10,
        workers: int = 1,
        use_multiprocessing: bool = False,
        return_dict: bool = False,
        **kwargs: Any,
    ) -> Union[float, List[float]]:
        """Return the loss value and metrics values for the model in test mode.

        Computation is done in batches (see the `batch_size`).

        Args:

        x: Input data.
            It could be:

            - A Numpy array (or array-like), or a list of arrays
                in case the model has multiple inputs.
            - A TensorFlow tensor, or a list of tensors
                in case the model has multiple inputs.
            - A dict mapping input names to the corresponding array/tensors
                if the model has named inputs.
            - A `tf.data` dataset.
                Should return a tuple
                of either `(inputs, targets)` or
                `(inputs, targets, sample_weights)`.
            - A `sarus.tensorflow.Dataset` dataset.
                Should return a tuple
                of either `(inputs, targets)` or
                `(inputs, targets, sample_weights)`.
            - A generator or `keras.utils.Sequence`
                returning `(inputs, targets)` or `(inputs, targets,
                sample_weights)`. A more detailed description of unpacking
                behavior for iterator types (Dataset, generator, Sequence)
                is given in the `Unpacking behavior for iterator-like inputs`
                section of `tensorflow.keras.Model.fit`.

        y: Target data.
            Like the input data `x`, it could be either Numpy
            array(s) or TensorFlow tensor(s). It should be consistent with
            `x` (you cannot have Numpy inputs and tensor targets, or
            inversely). If `x` is a dataset, generator or `keras.utils.Sequence`
            instance, `y` should not be specified (since targets
            will be obtained from the iterator/dataset).

        batch_size (int, optional):
            Number of samples per batch of
            computation. If unspecified, `batch_size` will default to 32. Do not
            specify the `batch_size` if your data is in the form of a dataset,
            generators, or `keras.utils.Sequence` instances (since they generate
            batches).

        verbose (int): 0 or 1.
            Verbosity mode. 0 = silent, 1 = progress bar.

        sample_weight (optional):
            Optional Numpy array of weights for the test samples,
            used for weighting the loss function. You can either pass a flat
            (1D) Numpy array with the same length as the input samples
            (1:1 mapping between weights and samples), or in the case of
            temporal data, you can pass a 2D array with shape `(samples,
            sequence_length)`, to apply a different weight to every timestep
            of every sample. This argument is not supported when `x` is a
            dataset, instead pass sample weights as the third element of `x`.

        steps (int, optional):
            Total number of steps (batches of samples)
            before declaring the evaluation round finished. Ignored with the
            default value of `None`. If x is a `tf.data` dataset and `steps` is
            None, `evaluate` will run until the dataset is exhausted. This
            argument is not supported with array inputs.

        callbacks: List of `keras.callbacks.Callback` instances.
            List of callbacks to apply during evaluation. See
            `Tensorflow callbacks documentation
            <https://www.tensorflow.org/api_docs/python/tf/keras/callbacks>`_.

        max_queue_size (int, optional):
            Used for generator or
            `keras.utils.Sequence` input only. Maximum size for the generator
            queue. If unspecified, `max_queue_size` will default to 10.

        workers (int, optional):
            Used for generator or `keras.utils.Sequence` input
            only. Maximum number of processes to spin up when using
            process-based threading. If unspecified, `workers` will default to
            1.

        use_multiprocessing (bool):
            Used for generator or `keras.utils.Sequence` input only. If `True`,
            use process-based threading. If unspecified, `use_multiprocessing`
            will default to `False`. Note that because this implementation
            relies on multiprocessing, you should not pass non-picklable
            arguments to the generator as they can't be passed easily to
            children processes.

        return_dict: If `True`, loss and metric results are returned as a
            dict, with each key being the name of the metric. If `False`, they
            are returned as a list.

        **kwargs:
            Unused at this time.

        See the discussion of `Unpacking behavior for iterator-like inputs` for
        `tensorflow.keras.Model.fit`.
        `Model.evaluate` is not yet supported with
        `tf.distribute.experimental.ParameterServerStrategy`.

        Returns:
            Scalar test loss (if the model has a single output and no metrics)
            or list of scalars (if the model has multiple outputs
            and/or metrics). The attribute `model.metrics_names` will give you
            the display labels for the scalar outputs.

        Raises:
            RuntimeError:
                If `model.evaluate` is wrapped in `tf.function`.

            ValueError:
                in case of invalid arguments.
        """
        if isinstance(x, Dataset):
            x = x._tensorflow()
        return super().evaluate(
            x=x,
            y=y,
            batch_size=batch_size,
            verbose=verbose,
            sample_weight=sample_weight,
            steps=steps,
            callbacks=callbacks,
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
            return_dict=return_dict,
            **kwargs,
        )

    def fit(
        self,
        x: Data = None,
        y: Data = None,
        batch_size: Optional[int] = None,
        epochs: Optional[int] = None,
        verbose: Union[int, str] = "auto",
        callbacks: List[tf.keras.callbacks.Callback] = None,
        validation_split: float = 0.0,
        validation_data=None,
        shuffle: Union[bool, str] = True,
        class_weight: Optional[Dict[int, float]] = None,
        sample_weight: Optional[np.ndarray] = None,
        initial_epoch: int = 0,
        steps_per_epoch: Optional[int] = None,
        validation_steps: Optional[int] = None,
        validation_batch_size: Optional[int] = None,
        validation_freq: Union[int, Container[int]] = 1,
        max_queue_size: int = 10,
        workers: int = 1,
        use_multiprocessing: bool = False,
        target_epsilon: Optional[float] = None,
        dp_l2_norm_clip: float = None,
        dp_noise_multiplier: float = None,
        dp_num_microbatches: Optional[int] = None,  # default: batch size
        **kwargs: Any,
    ) -> tf.keras.callbacks.History:
        """Train the model via the Sarus API.

        Args:

        x: The training data.
            It could be:

            - a sarus.tensorflow.Dataset
            - a tensorflow.data.Dataset

            If `target_epsilon` is greater than 0 then `x` must be a
            `sarus.tensorflow.Dataset`.

        y: Target data.
            Like the input data `x`,
            it could be either Numpy array(s) or TensorFlow tensor(s).
            It should be consistent with `x` (you cannot have Numpy inputs and
            tensor targets, or inversely). If `x` is a dataset, generator,
            or `keras.utils.Sequence` instance, `y` should
            not be specified (since targets will be obtained from `x`).
            Will be ignored if `target_epsilon` is specified.

        batch_size: Integer or `None`.
            Number of samples per gradient update.
            If unspecified, `batch_size` will default to 32.
            Do not specify the `batch_size` if your data is in the
            form of datasets, generators, or `keras.utils.Sequence` instances
            (since they generate batches).
            Will be ignored if `target_epsilon` is specified.

        epochs (int): Number of epochs to train the model.
            An epoch is an iteration over the entire data provided. Defaults to
            None. For remote DP-training (`target_epsilon` > 0), only one
            parameter among `epochs` and `dp_noise_multiplier` can be taken into
            account to satisfy `target_epsilon`. If both are set, the Sarus API
            will change `epochs`.

        verbose: 'auto', 0, 1, or 2. Verbosity mode.
            0 = silent, 1 = progress bar, 2 = one line per epoch.
            'auto' defaults to 1 for most cases, but 2 when used with
            `ParameterServerStrategy`. Note that the progress bar is not
            particularly useful when logged to a file, so verbose=2 is
            recommended when not running interactively (eg, in a production
            environment).

        callbacks: List of `keras.callbacks.Callback` instances.
            List of callbacks to apply during training.
            See `tf.keras.callbacks`. Note `tf.keras.callbacks.ProgbarLogger`
            and `tf.keras.callbacks.History` callbacks are created automatically
            and need not be passed into `model.fit`.
            `tf.keras.callbacks.ProgbarLogger` is created or not based on
            `verbose` argument to `model.fit`.
            Callbacks with batch-level calls are currently unsupported with
            `tf.distribute.experimental.ParameterServerStrategy`, and users are
            advised to implement epoch-level calls instead with an appropriate
            `steps_per_epoch` value.
            Will be ignored if `target_epsilon` is specified.

        validation_split: Float between 0 and 1.
            Fraction of the training data to be used as validation data.
            The model will set apart this fraction of the training data,
            will not train on it, and will evaluate
            the loss and any model metrics
            on this data at the end of each epoch.
            The validation data is selected from the last samples
            in the `x` and `y` data provided, before shuffling. This argument is
            not supported when `x` is a dataset, generator
            or `keras.utils.Sequence` instance.
            `validation_split` is not yet supported with
            `tf.distribute.experimental.ParameterServerStrategy`.
            Will be ignored if `target_epsilon` is specified.

        validation_data: Default to None.
            Data on which to evaluate the loss and any model
            metrics at the end of each epoch. The model will not be trained on
            this data. Thus, note the fact that the validation loss of data
            provided using validation_split or `validation_data` is not affected
            by regularization layers like noise and dropout. `validation_data`
            will override validation_split. `validation_data` could be:

            - a sarus.tensorflow.Dataset
            - a tensorflow.data.Dataset

            If `target_epsilon` is greater than 0 then `validation_data` must
            be None as it is not supported yet.

        shuffle: Boolean or str.
            Boolean (whether to shuffle the training data
            before each epoch) or str (for 'batch'). This argument is ignored
            when `x` is a generator or an object of tf.data.Dataset.
            'batch' is a special option for dealing
            with the limitations of HDF5 data; it shuffles in batch-sized
            chunks. Has no effect when `steps_per_epoch` is not `None`.
            Will be ignored if `target_epsilon` is specified.

        class_weight: Optional dict mapping class indices to a weight.
            Used for weighting the loss function
            (during training only).
            This can be useful to tell the model to
            "pay more attention" to samples from
            an under-represented class.
            Will be ignored if `target_epsilon` is specified.

        sample_weight: Optional Numpy array.
            Optional Numpy array of weights for
            the training samples, used for weighting the loss function
            (during training only). You can either pass a flat (1D)
            Numpy array with the same length as the input samples
            (1:1 mapping between weights and samples),
            or in the case of temporal data,
            you can pass a 2D array with shape `(samples, sequence_length)`,
            to apply a different weight to every timestep of every sample. This
            argument is not supported when `x` is a dataset, generator,
            or `keras.utils.Sequence` instance, instead provide the
            sample_weights as the third element of `x`.
            Will be ignored if `target_epsilon` is specified.

        initial_epoch: Integer.
            Epoch at which to start training
            (useful for resuming a previous training run).
            Will be ignored if `target_epsilon` is specified.

        steps_per_epoch: Integer or `None`.
            Total number of steps (batches of samples)
            before declaring one epoch finished and starting the
            next epoch. When training with input tensors such as
            TensorFlow data tensors, the default `None` is equal to
            the number of samples in your dataset divided by
            the batch size, or 1 if that cannot be determined. If x is a
            `tf.data` dataset, and 'steps_per_epoch'
            is None, the epoch will run until the input dataset is exhausted.
            When passing an infinitely repeating dataset, you must specify the
            `steps_per_epoch` argument. If `steps_per_epoch=-1` the training
            will run indefinitely with an infinitely repeating dataset.
            This argument is not supported with array inputs.
            When using `tf.distribute.experimental.ParameterServerStrategy`:
            * `steps_per_epoch=None` is not supported.
            Will be ignored if `target_epsilon` is specified.

        validation_steps: Optional Integer.
            Only relevant if `validation_data` is provided and
            is a `tf.data` dataset. Total number of steps (batches of
            samples) to draw before stopping when performing validation
            at the end of every epoch. If 'validation_steps' is None, validation
            will run until the `validation_data` dataset is exhausted. In the
            case of an infinitely repeated dataset, it will run into an
            infinite loop. If 'validation_steps' is specified and only part of
            the dataset will be consumed, the evaluation will start from the
            beginning of the dataset at each epoch. This ensures that the same
            validation samples are used every time.
            Will be ignored if `target_epsilon` is specified.

        validation_batch_size: Optional Integer.
            Number of samples per validation batch.
            If unspecified, will default to `batch_size`.
            Do not specify the `validation_batch_size` if your data is in the
            form of datasets, generators, or `keras.utils.Sequence` instances
            (since they generate batches).
            Will be ignored if `target_epsilon` is specified.

        validation_freq:  Integer or `collections.abc.Container` instance.
            Only relevant if validation data is provided.
            If an integer, specifies how many training epochs to run before a
            new validation run is performed, e.g. `validation_freq=2` runs
            validation every 2 epochs. If a Container, specifies the epochs on
            which to run validation, e.g. `validation_freq=[1, 2, 10]` runs
            validation at the end of the 1st, 2nd, and 10th epochs.
            Will be ignored if `target_epsilon` is specified.

        max_queue_size: Integer.
            Used for generator or `keras.utils.Sequence`
            input only. Maximum size for the generator queue.
            If unspecified, `max_queue_size` will default to 10.
            Will be ignored if `target_epsilon` is specified.

        workers: Integer.
            Used for generator or `keras.utils.Sequence` input
            only. Maximum number of processes to spin up
            when using process-based threading. If unspecified, `workers`
            will default to 1.
            Will be ignored if `target_epsilon` is specified.

        use_multiprocessing: Boolean.
            Used for generator or
            `keras.utils.Sequence` input only. If `True`, use process-based
            threading. If unspecified, `use_multiprocessing` will default to
            `False`. Note that because this implementation relies on
            multiprocessing, you should not pass non-picklable arguments to
            the generator as they can't be passed easily to children processes.
            Will be ignored if `target_epsilon` is specified.

        target_epsilon (Optional[float]): Defaults to None.
            Target epsilon for differentially private training.
            There are three behaviors depending on the value of `target_epsilon`:
              * `target_epsilon == 0` the training is performed locally on the
                synthetic data.
              * `target_epsilon > 0` training is performed remotely with
                 DP-SDG and only one parameter among `epochs` and
                 `dp_noise_multiplier` can be taken into account to satisfy
                 `target_epsilon`. If both are set, the Sarus API will
                 change `epochs`.
              * `target_epsilon is None` a default `target_epsilon` specific to
                 the current user and access rule is used.
                 Default target epsilon is 0 if the access is a Differentially-Private
                 access with per-user or per-group limit; default value equals to the
                 per-query limit if the access is a Differentially-Private access with
                 a per-query limit only. Meaning Sarus maximizes the result accuracy
                 in the given privacy constraints. See user documentation to know more.

        dp_l2_norm_clip (float): Defaults to None.
            Advanced DP-SGD parameter. The cumulative gradient across all
            network parameters from each microbatch is clipped so that its L2
            norm is at most this value. You should set this to something close
            to some percentile of what you expect the gradient from each
            microbatch to be. In previous experiments, we've found numbers from
            0.5 to 1.0 to work reasonably well.
            See `Tensorflow Privacy` documentation to know more.

        dp_noise_multiplier (float): Defaults to None.
            Advanced DP-SGD parameter. This governs the amount of noise added
            during training. Generally, more noise results in better privacy and
            lower utility. See `Tensorflow Privacy` documentation to know more.
            Note that for DP training (`target_epsilon`> 0), only one parameter
            among `epochs` and `dp_noise_multiplier` can be taken into account
            to satisfy `target_epsilon`. If both are set, the Sarus API will
            change `epochs`.

        dp_num_microbatches (int): Defaults to batch_size.
            Advanced DP-SGD parameter. The input data for each step (i.e.,
            batch) of your original training algorithm is split into this many
            microbatches. Generally, increasing this will improve your utility
            but slow down your training in terms of wall-clock time. The total
            number of examples consumed in one global step remains the same.
            This number should evenly divide your input batch size.
            See `Tensorflow Privacy` documentation to know more.
            If `None`, will default to `batch_size` of the dataset.

        Returns:
            History (tf.keras.callback.History):
                A history object containing training details.


        """
        if target_epsilon is None:
            client = x.dataset.client

            target_epsilon = client.session.get(
                f"{client.base_url}/datasets/{x.dataset.id}/default_epsilon"
            ).json()

        if target_epsilon < 0:
            raise ValueError(
                f"`target_epsilon` must be positive, got {target_epsilon}"
            )

        if target_epsilon == 0:
            if epochs is None:
                raise ValueError(
                    "The number of `epochs` should be provided when "
                    "`target_epsilon` is 0"
                )

            if dp_l2_norm_clip:
                logging.warning(
                    "Ignoring `dp_l2_norm_clip` with local training "
                    "as `target_epsilon` = 0. Specify `target_epsilon` > 0 "
                    "if you want a remote DP-training"
                )

            if dp_noise_multiplier:
                logging.warning(
                    "Ignoring `dp_noise_multiplier` with local training "
                    "as `target_epsilon` = 0. Specify `target_epsilon` > 0 "
                    "if you want a remote DP-training"
                )

            if dp_num_microbatches:
                logging.warning(
                    "Ignoring `dp_num_microbatches` with local training "
                    "as `target_epsilon` = 0. Specify `target_epsilon` > 0 "
                    "if you want a remote DP-training"
                )

            return self._fit_local(
                x=x,
                y=y,
                batch_size=batch_size,
                verbose=verbose,
                epochs=epochs,
                callbacks=callbacks,
                validation_split=validation_split,
                validation_data=validation_data,
                shuffle=shuffle,
                class_weight=class_weight,
                sample_weight=sample_weight,
                initial_epoch=initial_epoch,
                steps_per_epoch=steps_per_epoch,
                validation_steps=validation_steps,
                validation_batch_size=validation_batch_size,
                validation_freq=validation_freq,
                max_queue_size=max_queue_size,
                workers=workers,
                use_multiprocessing=use_multiprocessing,
                **kwargs,
            )
        else:
            if not isinstance(x, Dataset):
                raise TypeError(
                    "Expected `x` to be a sarus.tensorflow.Dataset when"
                    " `target_epsilon` > 0."
                )

            if x.original is True:
                raise ValueError(
                    "Training a model remotely using a "
                    "sarus.tensorflow.Dataset with `original`=True is not "
                    "supported yet."
                )

            if validation_data:
                logging.warning(
                    "Ignoring `validation_data` with remote training. "
                    "Not supported yet."
                )

            if epochs is not None and dp_noise_multiplier is not None:
                logging.warning(
                    "Only 1 parameter among `epochs` and `dp_noise_multiplier`"
                    " can be taken into account to satisfy `target_epsilon`. "
                    "Changing `epochs`."
                )

            return self._fit_remote(
                x=x,
                epochs=epochs,
                target_epsilon=target_epsilon,
                dp_l2_norm_clip=dp_l2_norm_clip,
                dp_noise_multiplier=dp_noise_multiplier,
                dp_num_microbatches=dp_num_microbatches,
                verbose=verbose,
            )

    def _fit_local(
        self,
        x: Data,
        y: Data,
        batch_size: Optional[int],
        verbose: Union[int, str],
        epochs: int,
        callbacks: List[tf.keras.callbacks.Callback],
        validation_split: float,
        validation_data: Data,
        shuffle: Union[bool, str],
        class_weight: Optional[Dict[int, float]],
        sample_weight: Optional[np.ndarray] = None,
        initial_epoch: int = 0,
        steps_per_epoch: Optional[int] = None,
        validation_steps: Optional[int] = None,
        validation_batch_size: Optional[int] = None,
        validation_freq: Union[int, Container[int]] = 1,
        max_queue_size: int = 10,
        workers: int = 1,
        use_multiprocessing: bool = False,
        **kwargs: Any,
    ) -> tf.keras.callbacks.History:
        if isinstance(x, Dataset):
            x = x._tensorflow()
            logging.info("Fitting model locally on synthetic data.")
        if isinstance(validation_data, Dataset):
            validation_data = validation_data._tensorflow()
        history = super().fit(
            x=x,
            y=y,
            batch_size=batch_size,
            verbose=verbose,
            epochs=epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            validation_data=validation_data,
            shuffle=shuffle,
            class_weight=class_weight,
            initial_epoch=initial_epoch,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            validation_batch_size=validation_batch_size,
            validation_freq=validation_freq,
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
            **kwargs,
        )
        return history

    def _fit_remote(
        self,
        x: Dataset,
        target_epsilon: float,
        dp_l2_norm_clip: Optional[float] = None,
        dp_noise_multiplier: Optional[float] = None,
        dp_num_microbatches: Optional[int] = None,  # default: batch size
        verbose: bool = True,
        epochs: Optional[int] = None,
        **kwargs: Any,
    ) -> tf.keras.callbacks.History:
        client = x.dataset.client
        if client is None:
            raise ValueError(
                f"The Sarus Dataset client is None: can not fit "
                f"remotely with `target_epsilon`={target_epsilon}."
            )

        batch_size, transforms = Model._refactor_transforms(
            x.dataset.transforms
        )

        if dp_num_microbatches and batch_size % dp_num_microbatches != 0:
            raise ValueError(
                f"`batch_size` should be a multiple of `dp_num_microbatches`, "
                f"got {batch_size} and {dp_num_microbatches}"
            )

        transform_def = Model._make_transform_def(transforms)

        # Set loss reduction to None for DP-SGD
        previous_reduction = self.loss.reduction
        self.loss.reduction = tf.keras.losses.Reduction.NONE
        if previous_reduction != tf.keras.losses.Reduction.NONE:
            logging.info("Setting losses.Reduction to NONE for DP-SGD.")

        try:
            # Build model weights
            self.predict(
                transform_def(
                    x.dataset._synthetic_as_tf_dataset(
                        batch_size=batch_size,
                        rows_number=None,
                        original=False,
                    )
                ).take(1)
            )

            # Push the model to the server
            serialized_model = Model._serialize_model(self)
            keras_model_resp = client.session.post(
                f"{client.base_url}/tasks_input_blobs",
                data=serialized_model,
            )
            if keras_model_resp.status_code > 200:
                raise Exception(
                    "Error while retrieving the model"
                    f"Full Gateway answer was:{keras_model_resp}"
                )
            keras_model_id = keras_model_resp.json()["id"]

            # Push the transform definition to the server
            # WARNING pickled functions require exactly same Python version b/w
            # sender and receiver (currently it is 3.6)
            if verbose:
                logging.info("Retrieving the preprocessing function")
            serialized_transform_def = cloudpickle.dumps(transform_def)
            transform_def_resp = client.session.post(
                f"{client.base_url}/tasks_input_blobs",
                data=serialized_transform_def,
            )
            if transform_def_resp.status_code > 200:
                raise Exception(
                    "Error while retrieving the preprocessing function"
                    f"Full Gateway answer was:{transform_def_resp}"
                )
            transform_def_id = transform_def_resp.json()["id"]

            # Launch the fit tasks
            request = client.session.post(
                f"{client.base_url}/fit",
                json={
                    "transform_def_id": transform_def_id,
                    "keras_model_id": keras_model_id,
                    "x_id": x.dataset.id,
                    "target_epsilon": target_epsilon,
                    "non_DP_training": False,
                    "batch_size": batch_size,
                    "dp_l2_norm_clip": dp_l2_norm_clip,
                    "dp_noise_multiplier": dp_noise_multiplier,
                    "dp_num_microbatches": dp_num_microbatches,
                    "seed": None,
                    "verbose": verbose,
                    "wait_for_completion": False,
                    "epochs": epochs,
                    **kwargs,
                },
            )
            if request.status_code > 200:
                if request.status_code == 403:
                    raise ValueError(
                        "Query failed with the following error: Privacy budget "
                        "limit exceeded"
                    )
                raise Exception(
                    f"Error while training the model.\
                                Full Gateway answer was:{request}"
                )
            task_id = request.json()["task"]
            if verbose:
                logging.info(f"Fitting task id: {task_id}")

            status = client._poll_training_status(task_id)
            error_message = status.get("error_message", None)
            if error_message:
                raise RuntimeError(
                    f"Training failed with the following error:\n"
                    f"{textwrap.indent(error_message, '  |')}"
                )

        finally:
            # Restore previous loss reduction
            self.loss.reduction = previous_reduction

        training_status = client._training_status(task_id)
        if "error_message" not in training_status:
            # Set fetched weights to model
            trained_model: tf.keras.Model = client._fetch_model(task_id)
            self.set_weights(trained_model.get_weights())

            # History
            history = tf.keras.callbacks.History()
            history.history = json.loads(
                training_status["result"].replace("'", '"')
            )
            if "qb_output_params" in training_status:
                history.params = json.loads(
                    training_status["qb_output_params"].replace("'", '"')
                )
            return history

    @staticmethod
    def _refactor_transforms(
        transforms: List[Transform],
    ) -> Tuple[int, List[Transform]]:
        """Refactor the `transforms`.

        Merge consecutive `batch` and `unbatch` operations for more efficient
        processign on the API side.

        Refactor the `split` transform to be applied on a batched dataset.

        NB: this could be removed once the API does not batch by default.
        """
        transform_types = [transform[0] for transform in transforms]

        if transform_types.count("batch") > 1:
            raise ValueError(
                "A Sarus Tensorflow Dataset can only be batched once "
                "for remote private training."
            )
        elif transform_types.count("batch") == 1:
            # The effective batch_size is the one defined in the batch operation
            batch_index = transform_types.index("batch")
            batch_transform = transforms[batch_index]
            batch_size = batch_transform[1]["batch_size"]
        else:
            # Unbatched dataset, rows will be retrieved one by one
            batch_size = 1

        if (
            "map" in transform_types
            and "batch" in transform_types
            and transform_types.index("map") < transform_types.index("batch")
        ):
            logging.warning(
                "A `map` has been applied before `batch`. "
                "The training performance will be degraded."
            )

        if len(transform_types) >= 2 and transform_types[:2] == [
            "unbatch",
            "batch",
        ]:
            # Safely merge batch and unbatch
            transforms = transforms[2:]
        elif len(transform_types) >= 3 and transform_types[:3] == [
            "unbatch",
            "split",
            "batch",
        ]:
            # Merge batch and unbatch but adapt split
            split_transform = transforms[1]
            start, end = (
                split_transform[1]["start"],
                split_transform[1]["end"],
            )
            new_split_transform = (
                "split",
                {"start": start // batch_size, "end": end // batch_size},
            )
            transforms = [new_split_transform] + transforms[3:]

        return batch_size, transforms

    @staticmethod
    def _make_transform_def(transforms: List[Transform]) -> Callable:
        def transform_def(
            ds: tf.data.Dataset, features: Optional[Dict] = None
        ) -> tf.data.Dataset:
            """Build a function to be sent remotely.

            This function should not make use of objects or functions defined
            in the Sarus module to avoid it being listed as a closure by
            cloudpickle.
            """
            func_mapping = {
                "map": lambda ds, params: ds.map(**params),
                "batch": lambda ds, params: ds.batch(**params),
                "unbatch": lambda ds, params: ds.unbatch(**params),
                "filter": lambda ds, params: ds.filter(**params),
            }
            for name, params in transforms:
                if name in func_mapping:
                    ds = func_mapping[name](ds, params)
                elif name == "split":
                    size = params["end"] - params["start"]
                    ds = ds.skip(params["start"]).take(size)
            return ds

        return transform_def

    @staticmethod
    def _serialize_model(model: tf.keras.Model) -> bytes:
        """Convert a keras Model to compressed archive format."""
        with tempfile.TemporaryDirectory() as _dir:
            logger = logging.getLogger("tensorflow")
            logger.setLevel(logging.WARNING)
            model.save(_dir)
            logger.setLevel(logging.INFO)
            with tempfile.TemporaryDirectory() as _second_dir:
                path = os.path.join(_second_dir, "tmpzip")
                with tarfile.open(path, mode="w:gz") as archive:
                    archive.add(_dir, recursive=True, arcname="")
                with open(path, "rb") as f:
                    ret = f.read()
                    return ret



"""Copyright 2020 Sarus SAS.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Sarus library to leverage sensitive data without revealing them

This lib contains classes and method to browse,
learn from & explore sensitive datasets.
It connects to a Sarus server, which acts as a gateway, to ensure no
results/analysis coming out of this lib are sensitive.
"""
from __future__ import annotations  # noqa: F407

import base64
import datetime
import decimal
import getpass
import hashlib
import io
import json
import logging
import os
import pprint
import re
import sys
import tarfile
import tempfile
import textwrap
import time
import typing as t
from types import TracebackType
import warnings
import webbrowser
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
import sarus_data_spec
from requests import Session
from sarus_data_spec.attribute import attach_properties

import sarus
import sarus.typing as srt
from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.manager.dataspec_api import (
    get_dataspec,
    pull_dataspec_graph,
    raise_response,
)
from sarus.manager.typing import SDKManager
from sarus.pandas.dataframe import DataFrame

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import sarus_data_spec.protobuf as sp
    import sarus_data_spec.typing as st
    from sarus_data_spec.constants import SCHEMA_TASK
    from sarus_data_spec.context import push_global_context
    from sarus_data_spec.dataspec_validator.privacy_limit import (
        DeltaEpsilonLimit,
    )
    from sarus_data_spec.protobuf.utilities import dict_deserialize
    from sarus_data_spec.status import DataSpecErrorStatus
    from sarus_data_spec.transform import select_sql
    from sarus_data_spec.variant_constraint import (
        pep_constraint,
    )

    from sarus.context.local_sdk import LocalSDKContext

    try:
        import tensorflow as tf

        import sarus.legacy.tensorflow as sltf
    except ModuleNotFoundError:
        pass  # error message in sarus_data_spec.typing


def sdk_excepthook(
    exc_cls: t.Type[BaseException],
    exc: BaseException,
    tb: Optional[TracebackType],
) -> Any:
    if isinstance(exc, DataSpecErrorStatus):
        last_exc_msg = exc.extract_last_exc()

        if last_exc_msg:
            sys.stderr.write("Remote dataspec error: ")
            sys.stderr.write(last_exc_msg)
            sys.stderr.write("\n")

            return

    sys.__excepthook__(exc_cls, exc, tb)


sys.excepthook = sdk_excepthook


class MessageToUserException(Exception):
    """An exception meant to print a message to the user when printed by the REPL after being raised."""

    args: t.Tuple[str]

    def __str__(self) -> str:
        indented_msg = "\n"
        indented_msg += textwrap.indent(self.args[0], "  ")

        return indented_msg


@dataclass
class Synthetic:
    """Simple class to represent synthetic data."""

    data: Optional[pa.Table] = None
    rows: Optional[int] = None


class DatasetList(list):
    """Custom class to lazily fetch a Dataset when indexing."""

    def __init__(
        self, iterable: t.Iterable[ShallowDataset], client: Client
    ) -> None:
        super().__init__(iterable)
        self.client = client

    def __getitem__(self, i: int) -> Dataset:
        shallow_dataset = super().__getitem__(i)
        slugname = shallow_dataset.name
        return self.client.dataset(slugname=slugname)

    def __setitem__(self, i: int, value: t.Any) -> None:
        raise NotImplementedError("A DatasetList is immutable.")

    def append(self, value: t.Any) -> None:
        raise NotImplementedError("A DatasetList is immutable.")

    def insert(self, i: int, value: t.Any) -> None:
        raise NotImplementedError("A DatasetList is immutable.")

    def extend(self, other: DatasetList) -> None:
        raise NotImplementedError("A DatasetList is immutable.")


class ShallowDataset:
    """Shallow dataset that is merely used as a placeholder for slugname."""

    def __init__(self, id: int, slugname: str):
        self.id = id
        self.name = slugname

    def __repr__(self) -> str:
        return f"<Sarus Dataset slugname={self.name} id={self.id}>"

    @classmethod
    def _from_dict(cls, data: dict) -> ShallowDataset:
        # `data` is the JSON returned by calling GET /datasets/
        return cls(data.get("id"), data.get("name"))


class Dataset(DataSpecWrapper[t.Iterator[pa.RecordBatch]]):
    """A class representing a Sarus Dataset.

    This class is the interface to the protected data. It enables to inspect the
    Sarus dataset metadata, manipulate synthetic data locally, prepare
    processing steps and identify the dataset for executing remote private
    queries.

    Args:
        id: The dataset id.
        client: The Sarus client where the dataset is defined.
        type_metadata: A serialized json holding the dataset metadata.
        marginals: A serialized json holding the dataset marginals.
        human_description: A short human readable description.
        policy:
    """

    def __init__(
        self,
        id: int,
        client: "Client",
        dataspec: st.DataSpec,
        is_bigdata: Optional[bool] = None,
        type_metadata: Optional[str] = None,
        human_description: Optional[str] = None,
        marginals: Optional[str] = None,
        policy: Optional[dict] = None,
        synthetic: Dict[str, Synthetic] = None,
    ):
        self.client: "Client" = client
        self._set_dataspec(dataspec)
        name = dataspec.properties().get("slugname")
        attach_properties(
            dataspec,
            name=srt.ATTRIBUTES_DS,
            properties={
                "id": str(id),
                "name": name,
                "desc": f"<Sarus Dataset slugname={name} id={str(id)}>",
            },
        )
        self.type_metadata: Optional[Dict[str, Any]] = None
        if type_metadata is not None:
            self.type_metadata = json.loads(type_metadata)
        self.human_description: Optional[str] = human_description
        if synthetic is None:
            self._synthetic = dict(original=Synthetic(), encoded=Synthetic())
        else:
            self._synthetic = synthetic
        self.marginals: Optional[Dict[str, Any]] = None
        if marginals is not None:
            self.marginals = json.loads(marginals)

        self.policy: Optional[dict] = policy

    @classmethod
    def _from_dict(cls, admin_ds: dict, client: Client) -> Dataset:
        """Get a dataset from the json data sent by the server.

        Deserializes the protobufs contained in the response and registers them
        in the storage.

        Args:
            admin_ds (dict): JSON data returned by calling GET /datasets/<id>
            client (Client): client used to get information from the server

        Returns:
            Dataset
        """
        dataspec_id = admin_ds.get("id")
        serialized_dataspec = admin_ds.get("dataspec")
        slugname = admin_ds.get("name")
        manager: SDKManager = client.context().manager()

        if serialized_dataspec is None:
            # TODO /datasets doesn't return the DataSpec def
            raise ValueError(f"{slugname} not returned by the server.")

        # Deserialize objects from the JSON response
        dataspec: st.Dataset = (
            client.context()
            .factory()
            .create(
                dict_deserialize(serialized_dataspec["dataset"]), store=False
            )
        )
        pull_dataspec_graph(client=client, uuid=dataspec.uuid())

        syn_dataset: st.Dataset = dataspec.variant(st.ConstraintKind.SYNTHETIC)

        # Update statuses
        dataspec.manager().copy_status_from_server(
            dataspec, task_names=[SCHEMA_TASK]
        )
        dataspec.manager().copy_status_from_server(
            syn_dataset, task_names=[SCHEMA_TASK]
        )

        # TODO fetch the PEP token from the API
        # (it is lazily computed on the server)
        h = hashlib.md5()
        h.update(dataspec.protobuf().SerializeToString())
        pep_token = h.hexdigest()
        pep_constraint(
            dataspec=dataspec,
            token=pep_token,
            required_context=[],
            privacy_limit=DeltaEpsilonLimit({0.0: 0.0}),
        )

        # Attach default delta to the source
        sources_ds = dataspec.sources(sp.type_name(sp.Dataset))
        source = sources_ds.pop()
        manager.set_admin_ds(source, admin_ds)

        return cls(
            id=dataspec_id,
            dataspec=dataspec,
            client=client,
            type_metadata=admin_ds.get("type_metadata"),
            marginals=admin_ds.get("marginals"),
            human_description=admin_ds.get("human_description"),
        )

    @property
    def features(self) -> Optional[Dict[str, Dict]]:
        """Features of the Sarus dataset and associated metadata.

        Returns:
            Dict[str, Dict]: A dictionary holding metadata where each key
            is a table name and each value is a dict with features.
        """
        if self.type_metadata is None:
            return None
        if len(self.type_metadata) > 1:
            features: Dict[str, Dict] = {
                table["name"]: table["features"]
                for table in self.type_metadata
            }
        else:
            features: Dict[str, Dict] = self.type_metadata[0]["features"]
        return features

    def __repr__(self) -> str:
        attributes_ds = self._dataspec.attribute(name=srt.ATTRIBUTES_DS)
        if attributes_ds is not None:
            return attributes_ds["desc"]
        else:
            return "<Sarus Dataset>"

    def _ipython_display_(self) -> None:
        """Custom ipython display."""
        print(repr(self))

    def as_tensorflow(
        self,
        max_download_size: Optional[int] = None,
        original: bool = False,
    ) -> sltf.Dataset:
        """Return the corresponding `sarus.tensorflow.Dataset` object.

        This allows to manipulate the Sarus `Dataset` as a Tensorflow dataset.

        Args:
            max_download_size (int, optional): Max number of synthetic data rows
                to download locally. Indicates the number of synthetic data
                rows to download from the Sarus server. It will not download
                more than the maximum number of available synthetic data. If
                `None`, it will download all the synthetic data. If different
                from `None`, all local computations will be done on the local
                synthetic sample so local results will differ from remote
                results.

            original (bool): Returns categories original values.
                If True will return categories as original values. If False,
                will encode categories as integers.

        Returns:
            A sarus_tensorflow.Dataset.
        """
        logging.warning(
            "'as_tensorflow()' uses the old API, not preprocessing."
        )
        if self._dataspec.manager().is_big_data(self._dataspec):
            raise Exception(
                "`as_tensorflow()` is not supported for this dataset"
            )
        return sltf.Dataset(
            self,
            max_download_size=max_download_size,
            original=original,
        )

    def as_pandas(self) -> DataFrame:
        attributes_ds = self._dataspec.attribute(name=srt.ATTRIBUTES_DS)
        assert attributes_ds is not None
        if self._dataspec.manager().is_big_data(self._dataspec):
            raise Exception("`as_pandas()` is not supported for this dataset")
        return DataFrame.from_dataspec(self._dataspec)

    def __getitem__(self, item: t.List[str]) -> Dataset:
        return self.table(item)

    def table(self, table_name: t.List[str]) -> Dataset:
        """Get a table from the dataset.

        Args:
            table_name (List[str]): Name of a form ['namespace_1','namespace_2'...,'table_name'].
                One can omit namespaces if table names are not ambiguous.

        Returns:
            Table: Table fitting the given name.
        """
        dataspec_tables = [
            el.to_strings_list()[0] for el in self._dataspec.schema().tables()
        ]
        # remove protection name and add schema_name
        dataspec_tables = [
            [self._dataspec.schema().name(), *(val for val in el[1:])]
            for el in dataspec_tables
        ]

        possible_tables_indices = []
        # iterate over tables and select_ones that have same path
        n_elements = len(table_name)
        for i, table_path in enumerate(dataspec_tables):
            if table_path[-n_elements:] == table_name:
                possible_tables_indices.append(i)

        try:
            assert len(possible_tables_indices) == 1
        except AssertionError:
            if len(possible_tables_indices) == 0:
                raise ValueError(f"Table {table_name} not found")
            else:
                raise ValueError(f"Ambiguous table name {table_name}")

        else:
            table_path = self._dataspec.schema().tables()[
                possible_tables_indices[0]
            ]

        ds_filter = self._dataspec.schema().data_type().get(table_path)
        filtered_ds = sarus_data_spec.transform.filter(filter=ds_filter)(
            self._dataspec
        )
        selected_ds = sarus_data_spec.transform.get_item(path=table_path)(
            filtered_ds
        )

        # create dataset
        table_ds = Dataset.from_dataspec(selected_ds)

        attributes_ds = self._dataspec.attribute(name=srt.ATTRIBUTES_DS)
        assert attributes_ds is not None
        name = f"<Sarus Table name={table_name}>"
        attach_properties(
            selected_ds,
            name=srt.ATTRIBUTES_DS,
            properties={
                "id": attributes_ds["id"],
                "name": name,
                "desc": f"<Sarus Table name={table_name}>",
            },
        )

        return table_ds

    def tables(self) -> t.List[t.List[str]]:
        """For given parameters of a Sarus Dataset return a list of Sarus Tables."""
        dataspec_tables = [
            el.to_strings_list()[0] for el in self._dataspec.schema().tables()
        ]

        # remove protection name and add schema_name
        dataspec_tables = [
            [self._dataspec.schema().name(), *(val for val in el[1:])]
            for el in dataspec_tables
        ]
        return dataspec_tables

    def sql(self, query: str) -> Dataset:
        """Apply an SQL query to the dataset.

        Args:
            query (str): SQL query

        Returns:
            an instance of Sarus Dataset.
        """
        ds = select_sql(query)(self._dataspec)
        attributes_ds = self._dataspec.attribute(name=srt.ATTRIBUTES_DS)
        assert attributes_ds is not None
        attach_properties(
            ds,
            name=srt.ATTRIBUTES_DS,
            properties={
                "id": attributes_ds["id"],
                "name": attributes_ds["name"],
                "desc": f"<Sarus Dataset query='{query}'>",
            },
        )
        ds_wrapper = self.from_dataspec(ds)
        return ds_wrapper

    @property
    def epsilon(self) -> float:
        """Retrieve the remaining global privacy budget (epsilon) of the current access rule.

        Returns:
            float: The remaining privacy budget (global epsilon) of the access
            rule.
        """
        attributes_ds = self._dataspec.attribute(name=srt.ATTRIBUTES_DS)
        assert attributes_ds is not None
        resp = self.client.session().get(
            f"{self.client.base_url}/datasets/{attributes_ds['id']}",
        )
        if resp.status_code > 200:
            raise Exception(
                f"Error while retrieving the current value of epsilon. "
                f"Gateway answer was: \n{resp}"
            )
        return float(resp.json()["accesses"][0]["current_epsilon"])

    @property
    def max_epsilon(self) -> float:
        """Retrieve the maximum global privacy budget (epsilon) granted by the Data preparator, for the current access rule.

        Returns:
            float: The maximum privacy budget (global epsilon) of the access
            rule.
        """
        attributes_ds = self._dataspec.attribute(name=srt.ATTRIBUTES_DS)
        assert attributes_ds is not None
        resp = self.client.session().get(
            f"{self.client.base_url}/datasets/{attributes_ds['id']}",
        )
        if resp.status_code > 200:
            raise Exception(
                f"Error while retrieving the current value of epsilon. "
                f"Gateway answer was: \n{resp}"
            )
        return float(resp.json()["accesses"][0]["max_epsilon"])

    def _fetch_synthetic(
        self,
        rows_number: Optional[int] = None,
        force_refresh: bool = False,
        original: bool = True,
        table_name: Optional[str] = None,
    ) -> pa.Table:
        """Fetch synthetic data as a pyarrow.Table.

        Downloads them if they are not in memory yet or if more rows are
        required. Do nothing if enough data have already been downloaded.

        Args:
            rows_number (int, optional): number of rows to return
            force_refresh (bool): if True, always fetch from server
            original (bool): if False, get categorical values as integers
            table_name (str, optional): The name of the table to fetch from.
        """
        if table_name is None and len(self.type_metadata) > 1:
            raise ValueError("Table name needs to be specified in this case.")

        dataset_size = int(self.marginals["rows"])
        if rows_number is None:
            rows_number = dataset_size

        if rows_number < dataset_size:
            logging.warning(
                "Requested `max_download_size` is lower than "
                "total number of remote synth data, so"
                "local operations' results will differ from remote results. \n"
                "Use `max_download_size`=None to locally download all "
                "available synth data"
            )

        if rows_number > dataset_size:
            logging.warning(
                "Cannot satisfy `max_download_size` as there is not enough "
                "remote synth data. \nDownloading max available synth data..."
            )

        rows_number = min(dataset_size, rows_number)

        synthetic_type = "original" if original else "encoded"
        has_enough_synthetic = (
            self._synthetic[synthetic_type].rows is not None
            and rows_number <= self._synthetic[synthetic_type].rows
        )

        if not has_enough_synthetic or force_refresh:
            # Fetch synthetic data
            attributes_ds = self._dataspec.attribute(name=srt.ATTRIBUTES_DS)
            uri = (
                f"{self.client.base_url}/synthetic_data/{attributes_ds['id']}"
            )
            if table_name is not None:
                uri = f"{uri}/{table_name}"

            resp = self.client.session().get(
                uri,
                stream=True,
                params={
                    "textual_categories": original,
                    "rows_number": rows_number,
                },
            )
            if resp.status_code > 200:
                raise Exception(
                    f"Error while retrieving synthetic data. "
                    f"Gateway answer was: \n{resp}"
                )

            synthetic_table = pq.ParquetFile(io.BytesIO(resp.content)).read()
            self._synthetic[synthetic_type].data = synthetic_table
            self._synthetic[synthetic_type].rows = rows_number

        return self._synthetic[synthetic_type].data.slice(length=rows_number)

    @staticmethod
    def _sarus_features_to_tf_spec(
        features: Dict[str, Any],
        original: bool,
    ) -> Dict[str, tf.TensorSpec]:
        """Convert Sarus features to Tensorflow spec."""
        mapping = {
            "categorical": tf.string if original else tf.int16,
            "boolean": tf.bool,
            "integer": tf.int64,
            "real": tf.float32,
            "text": tf.string,
            "datetime": tf.int64,
            "image": tf.uint8,
        }

        def get_tensorspec(feature) -> tf.TensorSpec:
            """Return the tensorflow spec of a Sarus feature."""
            feature_type = list(feature["type"].keys())[0]
            dtype = mapping[feature_type]
            if feature_type == "image":
                width = feature["type"]["image"]["shape"]["width"]
                height = feature["type"]["image"]["shape"]["height"]
                channels = feature["type"]["image"]["shape"]["channel"]
                shape = (None, width, height, channels)
            else:
                shape = (None,)
            return tf.TensorSpec(dtype=dtype, shape=shape)

        return {
            feature["name"]: get_tensorspec(feature) for feature in features
        }

    @staticmethod
    def _adapt_for_tf(
        batch: Dict[str, Any],
        features: Dict[str, Any],
        original: bool,
    ) -> Dict[str, Any]:
        """Convert python data to tensorflow compatible data.

        Convert datetime.datetime to nanoseconds.
        Convert images serialized as bytes by petastorm to 3d arrays.
        """
        adapted_batch = dict()
        for feature in features:
            dtype = list(feature["type"].keys())[0]
            if dtype == "datetime":

                def decode_datetime(datetime_value: datetime.datetime) -> int:
                    if datetime_value is None:
                        return np.iinfo(np.int64).min
                    return int(datetime_value.timestamp()) * int(1e9)

                adapted_batch[feature["name"]] = tf.constant(
                    list(map(decode_datetime, batch[feature["name"]])),
                    dtype=tf.int64,
                )
            elif dtype == "image":
                adapted_batch[feature["name"]] = tf.stack(
                    list(map(Dataset.decode_image, batch[feature["name"]])),
                    axis=0,
                )

            elif dtype == "text":
                adapted_batch[feature["name"]] = tf.constant(
                    list(map(str, batch[feature["name"]])), dtype=tf.string
                )

            elif dtype == "categorical":
                dtype = tf.string if original else tf.int16
                adapted_batch[feature["name"]] = tf.constant(
                    batch[feature["name"]], dtype=dtype
                )

            elif dtype == "integer":
                # replace None by min_int_32 to avoid crash
                cleaned_batch = [
                    x if x is not None else np.iinfo(np.int64).min
                    for x in batch[feature["name"]]
                ]
                adapted_batch[feature["name"]] = tf.constant(
                    cleaned_batch, dtype=tf.int64
                )

            elif dtype == "real":
                # replace None by NaN to avoid crash
                cleaned_batch = [
                    x if x is not None else float("NaN")
                    for x in batch[feature["name"]]
                ]
                adapted_batch[feature["name"]] = tf.constant(
                    cleaned_batch, dtype=tf.float32
                )

            elif dtype == "boolean":
                adapted_batch[feature["name"]] = tf.constant(
                    batch[feature["name"]], dtype=tf.bool
                )

        return adapted_batch

    def _synthetic_as_tf_dataset(
        self,
        batch_size: int,
        rows_number: Optional[int] = None,
        original: bool = False,
        force_refresh: bool = False,
        table_name: Optional[str] = None,
    ) -> tf.data.Dataset:
        """Return synthetic data as a tensorflow.data.Dataset.

        Args:
            batch_size (int): size of the batches in the dataset
            rows_number (int, optional): number of rows in the dataset
            original (bool): if False, return categories as integers
            force_refresh (bool): if True, always fetch from server

        Returns:
            tensorflow.data.Dataset: synthetic data
        """
        if table_name is None and len(self.type_metadata) > 1:
            raise ValueError("Table name needs to be specified in this case.")

        if force_refresh:
            self._fetch_synthetic(
                force_refresh=force_refresh,
                original=original,
                rows_number=rows_number,
            )

        if table_name is None:
            features = self.features
        else:
            features = self.features[table_name]

        # Generator function iterating pyarrow RecordBatches
        def generator() -> Dict[str, List[Any]]:
            synthetic_table = self._fetch_synthetic(
                original=original, rows_number=rows_number
            )
            for batch in synthetic_table.to_batches(max_chunksize=batch_size):
                yield Dataset._adapt_for_tf(
                    batch=batch.to_pydict(),
                    features=features,
                    original=original,
                )

        tf_signature = Dataset._sarus_features_to_tf_spec(features, original)

        return tf.data.Dataset.from_generator(
            generator, output_signature=tf_signature
        )


class Client:
    """Entry point for the Sarus API client."""

    _URL_SUFFIX = "/gateway"
    _HEADER_NAME_CLIENT_SDK_VERSION = "SARUS-Client-SDK-Version"

    def _url_validator(self, url):
        """URL validator.

        From https://stackoverflow.com/questions/7160737/
        python-how-to-validate-a-url-in-python-malformed-or-not
        """
        regex = re.compile(
            r"^(?:http|ftp)s?://"  # http:// or https://
            r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|"
            r"[A-Z0-9-]{2,}\.?)|"  # domain...
            r"localhost|"  # localhost...
            r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
            r"(?::\d+)?"  # optional port
            r"(?:/?|[/?]\S+)$",
            re.IGNORECASE,
        )
        return re.match(regex, url) is not None

    def __init__(
        self,
        url: str = f"http://0.0.0.0{_URL_SUFFIX}",
        google_login: bool = False,
        email: Optional[str] = None,
        password: Optional[str] = None,
        verbose: int = 1,
    ):
        # TODO : progress bar self.progress_bar = Progbar(100,
        # stateful_metrics=None)
        if not self._url_validator(url):
            raise ValueError("Bad url")
        if not url.endswith(self._URL_SUFFIX):
            raise ValueError(
                'Error: Invalid URL. '
                f'Please ensure you are using the correct URL endpoint ("{self._URL_SUFFIX}"). '
                'If you continue to experience issues, contact your Sarus account manager.'
            )
        self.base_url = url

        self._session = session = Session()
        session.headers[self._HEADER_NAME_CLIENT_SDK_VERSION] = sarus.VERSION

        if google_login:
            self._oidc_login()
        else:
            self._credentials_login(email, password)

        self._context = LocalSDKContext(self, verbose=verbose)
        self._context.set_sync_policy(srt.SyncPolicy.SEND_ON_VALUE)
        push_global_context(self._context)

    def set_verbose(self, verbose: int) -> None:
        self._context._verbose = verbose

    def context(self) -> LocalSDKContext:
        return self._context

    def url(self):
        return self.base_url

    def session(self):
        return self._session

    def save(self, wrapper: srt.DataSpecWrapper, path: str) -> None:
        """Save the representation of a sarus object to a file.

        Args:
            wrapper (DataSpecWrapper): The sarus object to save.
            path (str): The path to save the object to.

        Returns:
            None
        """
        manager: SDKManager = self.context().manager()

        user_ds = wrapper.dataspec(srt.DataSpecVariant.USER_DEFINED)
        manager.push(user_ds)

        saved_data = {
            "user_ds_uuid": user_ds.uuid(),
            "wrapper_type": type(wrapper).__name__,
        }

        with open(path, "w") as save_file:
            json.dump(saved_data, save_file)

    def load(
        self,
        path=None,
        uuid=None,
        to_dataset: bool = False,
    ) -> srt.DataSpecWrapper:
        """Load a sarus object from a path.

        Args:
            path (str): The file path to load the sarus object from.

        Returns:
            DataSpecWrapper: The sarus object (eg: DataFrame,model..).
        """
        if path is None and uuid is None:
            raise ValueError(
                "You must provide either argument 'path' or 'uuid'."
            )

        if uuid is not None:
            user_ds_uuid = uuid
            wrapper_type = None
        elif path is not None:
            with open(path, "r") as load_file:
                saved_data = json.load(load_file)

            user_ds_uuid = saved_data["user_ds_uuid"]
            wrapper_type = saved_data["wrapper_type"]

        pull_dataspec_graph(client=self, uuid=user_ds_uuid)
        ds = get_dataspec(client=self, uuid=user_ds_uuid)

        sources_ds = ds.sources(sp.type_name(sp.Dataset))
        for source in sources_ds:
            if not source.is_public():
                self.dataset(source["slugname"])

        if to_dataset and ds.prototype() == sp.Dataset:
            loaded_wrapper = Dataset.from_dataspec(ds)
        elif wrapper_type is not None and ds.prototype() == sp.Dataset:
            if wrapper_type == "Dataset":
                loaded_wrapper = Dataset.from_dataspec(ds)
            else:
                loaded_wrapper = DataFrame.from_dataspec(ds)
        else:
            # scalar must be created using the python type
            # transformed dataset also for now
            loaded_wrapper = self.context().wrapper_factory().create(ds)

        return loaded_wrapper

    def _oidc_login(self):
        oidc_login_url = f"{self.base_url}/oidc_login?headless=true"
        try:
            from IPython.display import Javascript, clear_output

            display(  # noqa: F821
                Javascript(f'window.open("{oidc_login_url}");')
            )
            display(clear_output())  # noqa: F821
        except Exception:
            webbrowser.open(oidc_login_url)
        token = getpass.getpass(
            "Logging in via google.\nYou will be redirected to a login page "
            "where you will obtain a token to paste below.\nIf you are not "
            f"redirected automatically, you can visit {oidc_login_url}\n"
        )
        self.session().cookies.set(
            "session", base64.b64decode(token).decode("ascii")
        )
        # just to check that the login is successful
        try:
            self.list_datasets()
        except Exception:
            raise Exception("Error during login: incorrect token")

    def _credentials_login(self, email=None, password=None):
        if email is None:
            raise ValueError("Please enter your email")

        credentials = {}
        credentials["email"] = email

        if password is not None:
            credentials["password"] = password
        else:
            credentials["password"] = getpass.getpass(
                prompt="Password: ", stream=None
            )

        response = self.session().post(
            f"{self.base_url}/login",
            json=credentials,
            headers={"Content-Type": "application/json"},
        )

        if response.status_code == 401:
            raise ValueError("Error during login: incorrect credentials")

        # Let `requests` handle unexpected HTTP status codes.
        response.raise_for_status()

        resp_json = response.json()
        warning_message = resp_json.get("warning_message")
        if warning_message:
            logging.warning(warning_message)

    def list_datasets(self) -> DatasetList:
        """List available Sarus datasets.

        Returns:
            DatasetList: List of available Sarus datasets.
        """
        request = self.session().get(f"{self.base_url}/datasets")
        raise_response(request)
        unsorted_datasets = [
            ShallowDataset._from_dict(ds_json)
            for ds_json in request.json()
            if ds_json["status"] == "ready"
        ]
        sorted_datasets = sorted(
            unsorted_datasets,
            key=lambda x: x.name,
        )
        return DatasetList(sorted_datasets, self)

    def dataset(self, slugname: str = None, id: int = None) -> Dataset:
        """Select a dataset from the Sarus Gateway.

        Either `slugname` or `id` should be provided. If both are provided, then
        only the `slugname` is considered.

        Args:
            slugname (str): the slugname of the Dataset to select.
            id (int): the id of the Dataset to select.

        Returns:
            Dataset: The selected Sarus dataset.

        """
        if slugname:
            return self._fetch_dataset_by_name(slugname)
        else:
            return self._fetch_dataset_by_id(id)

    def _fetch_dataset_by_id(self, id: int) -> Dataset:
        """Fetch a dataset from the Sarus Gateway.

        Args:
            id (int): id of the dataset to be fetched

        Returns:
            an instance of Dataset
        """
        request = self.session().get(f"{self.base_url}/datasets/{id}")
        raise_response(request)
        return Dataset._from_dict(request.json(), self)

    def _fetch_dataset_by_name(self, name: str) -> Dataset:
        """Fetch a dataset from the Sarus Gateway.

        Args:
            name (string): name of the dataset to be fetched

        Returns:
            Dataset: an instance of Dataset
        """
        request = self.session().get(
            f"{self.base_url}/datasets/name/{name}",
        )
        raise_response(request)
        dataset = Dataset._from_dict(request.json(), self)
        return dataset

    def query(
        self,
        query: str,
        target_epsilon: Optional[float] = None,
        verbose: bool = True,
        use_old_query=False,
        debug=False,
    ) -> int:
        """Execute a SQL query.

        Args:
            query (String): SQL query

            target_epsilon (Optional[float]): Maximum privacy budget (epsilon) to assign to the query.
                If 0, runs on the synthetic data.
                If >0, result is a combination of a query on the synthetic data and a
                Differentially-Private query on the real data.
                If None, a default target epsilon specific to the current user and access rule is used.
                Default target epsilon is 0 if the access is a Differentially-Private access with
                per-user or per-group limit; default value equals to per-query limit if the access is
                a Differentially-Private access with a per-query limit only. Meaning Sarus maximizes result
                accuracy in the given privacy constraints. See user documentation to know more.

            use_old_query (bool): Whether to use the v1 version of the query task.

            debug (bool): Print the remote traceback if available

        Returns:
            int: Id of the task.
        """
        endpoint = "query"
        if use_old_query:
            endpoint = "query_v1"

        payload = {
            "query": query,
        }
        if target_epsilon is not None:
            payload["target_epsilon"] = target_epsilon

        request = self.session().post(
            f"{self.base_url}/{endpoint}",
            json=payload,
        )
        if request.status_code > 200:
            if request.status_code == 403:
                raise ValueError(
                    "Query failed with the following error: Privacy budget "
                    "limit exceeded"
                )
            else:
                try:
                    error_message = request.json()["error_message"]
                    message_to_user = error_message

                    error_chain = request.json().get("error_chain", None)
                    if error_chain is not None and debug:
                        message_to_user += "\n\nRemote Traceback:\n\n"
                        message_to_user += pprint.pformat(error_chain)
                    raise MessageToUserException(message_to_user)
                except (json.JSONDecodeError, KeyError):
                    raise Exception(
                        f"Error while sending a query.\
                                                             Full Gateway answer was:{request}"
                    )

        task_id = request.json()["task"]
        dataset_id = request.json()["dataset"]["id"]
        start_eps = request.json()["dataset"]["accesses"][0]["current_epsilon"]
        status = self._poll_query_status(task_id)
        error_message = status.get("error_message", None)
        error_chain = status.get("error_chain", None)

        if error_message is not None:
            message_to_user = error_message

            if error_chain is not None and debug:
                message_to_user += "\n\nRemote Traceback:\n\n"
                message_to_user += pprint.pformat(error_chain)
            raise MessageToUserException(message_to_user)
        if verbose:
            ds = self._fetch_dataset_by_id(dataset_id)
            logging.info(
                f"Actual privacy consumption (epsilon): "
                f"{ds.epsilon-start_eps:.03f}"
            )

        if status["status"] == "SUCCESS":
            warning_msg = status["warning_message"]
            if warning_msg is not None:
                logging.warning(warning_msg)

            status_with_pyobjs = self._convert_to_result_with_pyobjs(status)

            return status_with_pyobjs
        else:
            return status

    @staticmethod
    def _convert_to_result_with_pyobjs(
        status: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Convert the query result values to native Python objects using the description field.

        Args:
            status (dict): the status response for a given completed query task returned by the server.

        Returns:
            dict: a new status dict with the initialized Python objects.
        """
        result = []
        for row in status["result"]:
            for idx, col_desc in enumerate(status["description"]):
                _, class_name, _, _, _, _, _ = col_desc
                if class_name == str(decimal.Decimal):
                    row[idx] = decimal.Decimal(row[idx])
            result.append(row)
        return {
            **status,
            "result": result,
        }

    def _abort_training(self, id: int):
        """Abort a training on the Sarus Gateway.

        Args:
            id (int): id of the task to abort (provided by the fit method).
        """
        resp = self.session().delete(
            f"{self.base_url}/training_tasks/{id}/abort",
        )
        if resp.status_code != 204:
            raise Exception(
                f"Error while trying to abort task:\n{resp.content}"
            )

    def _training_status(self, id: int) -> dict:
        """Fetch a dataset from the Sarus Gateway.

        Args:
            id (int): id of the task to be queried. It was provided by the fit
            method

        Returns:
            dict: a dict with the status of a training tasks
        """
        request = self.session().get(
            f"{self.base_url}/training_tasks/{id}",
        )
        return request.json()

    def _poll_training_status(self, id: int, timeout: int = 1000) -> dict:
        """Poll & display the status of a training task.

        Args:
            id (int): id of the task to be queried. It was provided by the fit
            method

            timeout (int): in seconds

        Returns:
            dict: The training status at the end of the task

        Raises:
            TimeoutError: if timeout is reached before the training finishes
        """
        offset = 0
        elapsed_time = 0.0
        while elapsed_time < timeout:
            elapsed_time += 0.5
            request = self.session().get(
                f"{self.base_url}/training_tasks/{id}",
                params=dict(offset=offset),
            )
            response_dict = request.json()
            offset = response_dict.get("next_offset", 0)
            if "progress" in response_dict:
                progress = base64.b64decode(
                    response_dict["progress"].encode("ascii")
                ).decode("ascii")
                if progress:
                    sys.stdout.write(progress)
            else:
                # this is the end of the training
                sys.stdout.write("\n")
                return response_dict
            sys.stdout.flush()
            time.sleep(0.5)
        raise TimeoutError(
            "Timeout reached while waiting for the model training to finish."
        )

    def _poll_query_status(self, id: int, timeout: int = 1000) -> dict:
        """Poll & display the status of a query task.

        Args:
            id (int): id of the task to be queried. It was provided by the fit
            method

            timeout (int): in seconds

        Returns:
            dict: The query status at the end of the task

        Raises:
            TimeoutError: if timeout is reached before the query finishes
        """
        offset = 0
        elapsed_time = 0.0
        while elapsed_time < timeout:
            elapsed_time += 0.5
            request = self.session().get(
                f"{self.base_url}/query_tasks/{id}",
                params=dict(offset=offset),
            )
            response_dict = request.json()
            status = response_dict["status"]
            if status != "PENDING":
                return response_dict
            time.sleep(0.5)
        raise TimeoutError(
            "Timeout reached while waiting for the model training to finish."
        )

    def _fetch_model(self, id: int) -> tf.keras.Model:
        """Fetch a trained model from the Sarus Gateway.

        Args:
            id (int): id of the task to be queried. It was provided by the fit
                method

        Returns:
            tf.keras.Model: a Keras model
        """
        response = self.session().get(
            f"{self.base_url}/models/{id}",
        )
        # apparently we need to save to a temp file
        # https://github.com/keras-team/keras/issues/9343
        with tempfile.TemporaryDirectory() as _dir:
            f = tarfile.open(fileobj=io.BytesIO(response.content), mode="r:gz")
            f.extractall(_dir, members=None)

            return tf.keras.models.load_model(_dir)


def _save_model(model: tf.keras.Model) -> bytes:
    """Convert a keras Model to compressed archive format."""
    with tempfile.TemporaryDirectory() as _dir:
        model.save(_dir)
        with tempfile.TemporaryDirectory() as _second_dir:
            path = os.path.join(_second_dir, "tmpzip")
            with tarfile.open(path, mode="w:gz") as archive:
                archive.add(_dir, recursive=True, arcname="")
            with open(path, "rb") as f:
                ret = f.read()
                return ret



from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import sklearn.compose as sk_compose
    from sklearn.compose import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class ColumnTransformer(DataSpecWrapper[sk_compose.ColumnTransformer]):
    @sarus_init("sklearn.SK_COLUMN_TRANSFORMER")
    def __init__(
        self,
        transformers,
        *,
        remainder="drop",
        sparse_threshold=0.3,
        n_jobs=None,
        transformer_weights=None,
        verbose=False,
        verbose_feature_names_out=True,
    ):
        ...

    @sarus_method("sklearn.SK_PIPELINE_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()
